<copyright> TESLA mu.BlockDefinition extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1999 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: BlockDefinition.t,v 1.15 2001/04/08 22:39:48 bruce Exp $</id>
    </copyright>

/******************** BlockDefinition ********************/

implementation class
BlockDefinition extension tesla
{
  <doc> Block counter, used for creating unique block names.  </doc>
  static int next_block;

  <doc> The name parts of each block is the same, {eval}.  </doc>
  static Indexed eval_name_parts;
}

end;

implementation instance
BlockDefinition extension tesla
{
  <doc> The number of this block.  </doc>
  int block_id;

  <doc> The body of this block.  </doc>
  Compound body_expression;

  <doc> The containing compound for this block.  It basically contains the
      an assignment of the {body} to the return value of this block.  </doc>
  BlockBody body;

  <doc> The e-val-ue type.  </doc>
  Type eval_type;

  <doc> The variables that will hold the eval value of this block.  </doc>
  public ArgumentTuple eval_names;

  <doc> Variable proxies, if employed by this block.  </doc>
  MutableDictionary var_context;

  <doc> When set, {self} is needed in this block.  </doc>
  public VariableSingle nameof_self;
}

redefine id
  initWithContainer Context c
{
  block_id = next_block++;

  argument_singles = [MutableDictionary new];
  ArgumentSingle a = [[ArgumentSingle alloc]
		      initWithName [UniqueString with "self-block"]
		      type basic_types["id"]];
  argument_singles[[a name]] = a;
  a = [[ArgumentSingle alloc] initWithName [UniqueString with "cmd-block"]
			      type basic_types["selector"]];
  argument_singles[[a name]] = a;

  = [super initWithContainer c];
}

<doc> Return {self}.  </doc>
Callable
  callable
{
  = self;
}

<doc> Return the {var_context}.  </doc>
Mapped
  context
{
  = var_context;
}

void
  finishBlock
{
}

<doc> Return the {eval_name_parts}, creating them upon demand.  </doc>
Indexed
  name_parts
{
  if (!eval_name_parts)
    eval_name_parts = [ObjectArray with "eval"];
  = eval_name_parts;
}

void
  needSelf VariableSingle ns
{
  nameof_self = ns;
}

void
  precompileDeclarations Cblock parent
		  method Cblock body
{
  [super (Callable) precompileDeclarations parent method body];
}

Compound
  startBlockBody
pre
  !body_expression
{
  body = [[BlockBody alloc] initWithContainer self];
  body_expression = [[Compound alloc] initWithContainer body];
  [Compound (container) registerBlock self];
  = body_expression;
}

void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  [body coarseUsage f parent self body m];
  if (block_variables != nil)
    {
      Enumerator e = [block_variables enumerator];
      BlockVariable v;
      boolean bv;

      while ({(bv, v) = [e next]; bv;})
        [v coarseUsage f parent self body m];
    }
  [super coarseUsage f parent c body m];
}

TeslaOutput (of)
  compileBlocks TeslaOutput of
{
  [formal_selector noteUsage];
  [body compileBlocks of];

  [[of nl self] print ("/* start of block ", self, " */")];

  if (block_variables != nil)
    {
      int num;

      [self writeBlockStructName [[of nl self] print "struct "]];
      [[of nl self] print "{"];
      [of indent 2];

      Enumerator e = [block_variables enumerator];
      BlockVariable v;
      boolean bv;

      while ({(bv, v) = [e next]; bv;})
	{
	  num++;
	  [[v writeCDeclaration [of nl]] print ';'];
	}

      [of indent -2];
      [[of nl self] print "};"];

      [[self writeBlockDescriptionName [[of nl self]
					 print "struct trtd_block "]]
	print (" = {", num, ", {")];

      e = [block_variables enumerator];
      while ({(bv, v) = [e next]; bv;})
	[v writeCNormalDescription of];

      [of print "}};"];
    }
      
  [[of nl self] print "static "];
  
  [[eval_type tupleFirstElement] writeCReturnType of];
  [[self writeBlockCodeName [of nl self]] print " ("];
  [[Meta (type) writeCReferenceTypeName of]
    print " block_self, selector block_cmd"];

  if (arguments != nil)
    {
      int i, n = [arguments length];
      for (i = 0; i < n; i++)
	{
	  Argument a = arguments[i];
	  [a writeCDeclaration of prefix ", "];
	}
    }
  if (eval_names != nil)
    [eval_names writeCFile of prefix ", " infix " *_r"
		postfix nil skipFirst YES];

  [of print ')'];

  Cclosure b = [[Cclosure alloc] initWithBlock self];
  [body precompileElements b method b];
  [b writeCFile of];

  [[of print ("\n/* end of block ", self, " */")] nl];
}

<doc> Pass to the enclosing method.  </doc>
long (result)
  contextID
{
  = [[[self container] method] contextID];
}

<doc> Return {TRUE} if we have {block_variables}.  </doc>
boolean
  haveBlockVariables
{
  = block_variables != nil;
}

Cexpr (result)
  precompile Cblock parent
      method Cblock method_body
{
  result = [block_instance temporaryCexpr parent];

  if (block_variables != nil)
    {
      Enumerator e = [block_variables enumerator];
      BlockVariable v;
      boolean bv;

      /* Allocate space for the block variables.  */
      MutableByteString ms = [MutableByteString new];
      [[self writeBlockName ms] print " = xmalloc (sizeof *"];
      [[self writeBlockName ms] print ")"];
      [parent add [[Cstmt alloc self] init ms]];

      /* Initialize the block variables.  */
      while ({(bv, v) = [e next]; bv;})
	{
	  Cref r = [v precompile parent method method_body];
	  [[[v initial_value] precompile parent method method_body]
	    compile parent assignmentTo r];
	}
    }

  [[[Cmkblock alloc] initWithBlock self] compile parent assignmentTo result];
}

(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (!type)
    {
      type = block_instance;

      [body_expression typeAndResolve];
      eval_type = [body_expression type];

      if (block_variables != nil)
        {
          Enumerator e = [block_variables enumerator];
          BlockVariable v;
          boolean bv;

          while ({(bv, v) = [e next]; bv;})
            [v resolveInitialValue context_type];
        }

      [self formal_selector];

      /* Create variables for the return values.  Add the body to the body
         container.  */
      if (eval_type == void_type)
	[body addExpression body_expression];
      else
	{
	  eval_names = [ArgumentTuple withVariableTuple
			[eval_type temporaryTuple body]];
	  [eval_names updateLocation self];
	  Assignment a = [[Assignment alloc self]
			   initWithLhs [eval_names reference nil]
			   rhs body_expression];
	  [body addExpression a];
	}
    }

  = [super resolve (expected, convertible) context context_type
	   nested: (indices, index)];
}

<doc> Return the {eval_type}.  </doc>
Type
  return_type
{
  = eval_type;
}

<doc> If needed, create a proxy for the variable {v} and add it to the
    {var_context}.  Return the proxy for {v} from {var_context}.  The {mu}
    implementation returns {v}.  </doc>
VariableBlockProxy (w)
  variableProxy LocalVariable v
{
  if (!var_context)
    var_context = [MutableDictionary new];
  else
    w = var_context[v];

  if (!w)
    {
      w = [[VariableBlockProxy alloc] initWithVariable v fromBlock self];
      var_context[w] = v;
    }
}

OutputStream (s)
  writeBlockName OutputStream s
{
  [s print ("_block_", block_id)];
}

OutputStream (s)
  writeBlockDescriptionName OutputStream s
{
  [s print ("_bd_block_", block_id)];
}

OutputStream (s)
  writeBlockCodeName OutputStream s
{
  [s print ("_code_block_", block_id)];
}

OutputStream (s)
  writeBlockStructName OutputStream s
{
  [s print ("_block_", block_id)];
}

OutputStream (s)
  writeContextName OutputStream s
{
  [s print ("_context_", [self contextID])];
}

OutputStream (s)
  writeContextStructName OutputStream s
{
  [s print ("_context_", [self contextID])];
}

end;
