<copyright> TESLA mu.Expression extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Expression.t,v 1.24 2000/08/27 15:26:51 tiggr Exp $</id>
    </copyright>

/******************** Expression ********************/

implementation class
Expression extension tesla: tesla
{
  <doc> Various flags.  {EXPR_RHS} is set if the value of the expression
      is actually used.  {EXPR_LHS} is set if the expression is the target
      of an assignment.  {EXPR_UNWIND} is set in the protection expression
      of an {unwind}.  {EXPR_BIND} is set in the handlers of a {bind}.
      </doc>
  const EXPR_RHS = 1;
  const EXPR_LHS = 2;
  const EXPR_BIND = 4;
  const EXPR_UNWIND = 8;
}

end;

implementation instance
Expression extension tesla
{
  <doc> Flags.  </doc>
  int expr_flags;
}

<doc> Note the usage of the receiving expression.  Notable flags are
    {EXPR_RHS} and {EXPR_LHS}.  </doc>
void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  if (f & EXPR_RHS != 0)
    expr_flags |= EXPR_RHS;
  if (f & EXPR_LHS != 0)
    expr_flags |= EXPR_LHS;
  if (f & EXPR_BIND != 0)
    expr_flags |= EXPR_BIND;
  if (f & EXPR_UNWIND != 0)
    expr_flags |= EXPR_UNWIND;
}

<doc> Before a method is compiled, any blocks that it contains must be
    compiled first.  The {compileBlocks} methods are used to traverse all
    compounds; the blocks will compile themselves.  When we have blocks,
    special-purpose traversal methods like this will no longer be
    necessary :-)  </doc>
TeslaOutput (of)
  compileBlocks TeslaOutput of
{
}

<doc> Return {TRUE} iff the receiving expression is fully constant.
    Default returns {FALSE}.  </doc>
boolean
  isConstant
{
}

Cexpr (result)
  precompile Cblock parent
      method Cblock body
{
  if (self == void_expr)
    result = [[Cstmt alloc] init "void"];
  else
    {
      [[[ParseTom error self] print ("unimplemented: ", cmd, " ", self)] nl];
      result = [[Cstmt alloc] init "UNIMPLEMENTED PRECOMPILE"];
      [parent add result];
    }
}

Cexpr (result)
    precompile Cblock parent
	method Cblock body
  assignmentOf Cexpr rhs
pre
  rhs != nil
post
  result != nil
{
  [[[ParseTom error self] print ("invalid lhs: ", self)] nl];
}

Cexpr (result)
    precompile Cblock parent
	method Cblock body
  assignmentTo Expression lhs
pre
  lhs != nil
post
  result != nil
{
  = [lhs precompile parent method body
	 assignmentOf [self precompile parent method body]];
}

<doc> Type the receiving expression and return it resolved.  </doc>
Expression (result)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index) = (nil, -1)
pre
  ((!expected || !convertible)
   && (index < 0 == !indices))
{
  /* This is (toplev.m)resolve_expr from tomc.  */
  Cons possible_types;
  (result, possible_types) = [self resolve (expected, convertible)
				   context context_type
				   nested: (indices, index)];

  if (possible_types != nil && ![result type])
    /* The expression can have _some_ type but it can not decide.
       Pick the first one from the POSSIBLE_TYPES.  */
    (result, possible_types)
      = [result resolve ([Cons with ([possible_types car], nil)], nil)
		context context_type nested: (indices, index)];

  if (!possible_types || ![result type])
    {
      [[[ParseTom error self] print "unable to deduce type of expression"] nl];
      if (possible_types != nil)
	[[[ParseTom report self]
	   print ("with possible types: ", possible_types)] nl];
      if (expected != nil)
	[[[ParseTom report self]
	   print ("in context of possible types: ", expected)] nl];
      else if (convertible != nil)
	[[[ParseTom report self]
	   print ("in context of cast to: ", convertible)] nl];
      else
	[[[ParseTom report self] print ("in void context")] nl];

      [[[ParseTom report self]
	 print ("problem caused by the following expression: ", self)] nl];
    }
}

<doc> Resolve the last expression in the receiving compound.  The default
    implementation simply resolves {self}.  </doc>
Expression (result)
  resolveLastExpression (Cons, Type) (expected, convertible)
		context Meta context_type
		nested: (MutableIntArray, int) (indices, index) = (nil, -1)
{
  = [self resolve (expected, convertible) context context_type
	  nested: (indices, index)];
}

<doc> The one and only resolution method.  Subclasses should override
    this.

    If {expected == nil}, the {possible_types} must be a list of possible
    types of the receiving expression.  If {expected != nil}, the
    {possible_types} must be a subset of {expected}, in the same order.

    If {index == -1}, the receiving expression must resolve against the
    {expected} or {convertible} types.  If {index >= 0}, the numbers in
    {indices}, up to {index}, contain the indices into the type tuples in
    {expected} or {convertible}, down to the type applicable for the
    receiving expression.  </doc>
// tomCCC does not like this...
// Mon May 11 21:28:28 1998, tiggr@gerbil.org
//   ((!expected || !convertible)
//    && (index >= 0 -> indices != nil))
(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index) = (nil, -1)
pre
  (!expected || !convertible)
{
  if (!type)
    {
      [[[ParseTom error self] print ("internal: ", cmd, " with no type")] nl];
      [[[ParseTom report self] print ("offending object: ", self)] nl];
      if (expected != nil)
	[[[ParseTom report self] print ("expected: ", expected)] nl];
      return;
    }

  /* This implementation is (toplev.m)default_resolve_expr from tomc.  */
  Type ast = [type actualSelf context_type];
  if (!expected)
    return (self, [Cons with (ast, nil)]);

  Cons last;
  Type s, t;

  while (expected != nil)
    {
      (t, expected) = [expected decons];

      // Wrong: the last element of an argument tuple can be dynamic...
      // Mon Apr 27 23:59:03 1998, tiggr@gerbil.org
      s = (t == dynamic_type) ? ast : [t typeAt index in indices];
      s = [s actualSelf context_type];
      if (s != nil && [ast distanceToFormal s] >= 0)
	{
	  Cons c = [Cons with (t, nil)];
	  if (!possible_types)
	    possible_types = c;
	  else
	    [last set_cdr c];
	  last = c;
	}
    }
  resolved = self;
}

<doc> Return a replacement for the {assignment}, of which the receiver is
    the lhs, or return {self}.  The default implementation returns {self}.
    </doc>
Expression
  replacementForAssignmentBy Assignment assignment
{
  = self;
}

<doc> Return the first element of the receiving tuple of....  The default
    implementation returns {self}.  </doc>
Expression
  tupleFirstElement
{
  = self;
}

<doc> Resolve the receiving expression in a void context.  It must not
    change its identity.  </doc>
void
  typeAndResolve
{
  Expression new_self = [self resolve (nil, nil) context current_meta];
  if (new_self != self)
    {
      [[[ParseTom error self] print ("internal: resolving ", self)] nl];
      [[[ParseTom report self] print ("resulted in ", new_self)] nl];
    }
}

<doc> Constant expressions know how to write themselves to a C output
    file.  Others do not.  </doc>
deferred OutputStream (s)
  writeCFile OutputStream s;

end;

/******************** AnyExpression ********************/

implementation class
AnyExpression extension tesla: tesla

end;

implementation instance
AnyExpression extension tesla

Cexpr (result)
    precompile Cblock parent
	method Cblock body
  assignmentOf Cexpr rhs
{
  = rhs;
}

end;

/******************** ExpressionTuple ********************/

implementation class
ExpressionTuple extension tesla: tesla

end;

implementation instance
ExpressionTuple extension tesla

<doc> Pass to the elements.  </doc>
void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  int i;
  for (i = 0; i < length; i++)
    [self[i] coarseUsage f parent c body m];
}

<doc> Inform our elements.  </doc>
TeslaOutput (of)
  compileBlocks TeslaOutput of
{
  int i;
  for (i = 0; i < length; i++)
    [self[i] compileBlocks of];
}

Cexpr
  precompile Cblock parent
      method Cblock body
{
  Ctuple result = [[Ctuple alloc self] init length];
  int i;
  for (i = 0; i < length; i++)
    [result add [self[i] precompile parent method body]];
  = result;
}

Cexpr
    precompile Cblock parent
	method Cblock body
  assignmentOf Cexpr rhs
{
  Ctuple ct = [Ctuple (rhs) temporaries parent method body];
  Ctuple result = [[Ctuple alloc self] init length];
  int i;
  for (i = 0; i < length; i++)
    [result add [self[i] precompile parent method body assignmentOf ct[i]]];
  = result;
}

(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (type != nil)
    return [super resolve (expected, convertible) context context_type
		  nested: (indices, index)];

  boolean is_dynamic;
  int i;

  if (expected != nil)
    {
      /* Drop all types from EXPTECTED that do not match this tuple with
         respect to number of elements.  */
      Cons list, last, l = expected;
      Type t;

      while (l != nil)
	{
	  (t, l) = [l decons];
	  Type tt = [t typeAt index in indices];
	  if (tt == dynamic_type || [tt isTupleOf length])
	    {
	      if (tt == dynamic_type)
		is_dynamic = TRUE;
	      Cons c = [Cons with (t, nil)];
	      if (!list)
		list = c;
	      else
		[last set_cdr c];
	      last = c;
	    }
	}
      if (!list)
	return (self, nil);
      expected = list;
    }

  boolean fully_typed;
  MutableIntArray our_indices = !indices ? [MutableIntArray new] : indices;
  int our_index = index + 1;
  for (;;)
    {
      fully_typed = TRUE;

      /* Resolve the elements.  */
      for (i = 0; i < length; i++)
	{
	  Expression f, e = self[i];
	  Cons l;

	  if (!e)
	    continue;

	  if (expected != nil)
	    {
	      our_indices[our_index] = i;
	      (f, l) = [e resolve (expected, nil) context context_type
			  nested: (our_indices, our_index)];
	    }
	  else
	    (f, l) = [e resolve (nil, nil) context context_type];

	  if (e != f)
	    self[i] = e = f;

	  if (!l)
	    return (self, nil);
	  if (![e type])
	    fully_typed = FALSE;

	  if (expected != nil)
	    {
	      if (![l equal expected])
		{
		  expected = l;

		  /* Restart at the beginning.  */
		  if (i > 0)
		    break;
		}
	    }
	}
      if (i == length)
	break;
    }

  if (fully_typed)
    {
      TypeTuple tt = [TypeTuple withCapacity length];
      for (i = 0; i < length; i++)
	{
	  Expression e = self[i];
	  if (e != nil)
	    [tt add [[e type] actualSelf context_type]];
	}
      type = tt;

      return [super resolve (expected, convertible) context context_type
		    nested: (indices, index)];
    }

  if (expected != nil)
    return (self, expected);

  [self unimplemented cmd message: "!fully_typed && !expected"];
}

<doc> Pass to the first element.  </doc>
Expression
  tupleFirstElement
{
  = [self[0] tupleFirstElement];
}

end;
