<copyright> TESLA mu.Meta extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Meta.t,v 1.24 2001/04/08 22:39:48 bruce Exp $</id>
    </copyright>

implementation class
Meta: tesla, posing mu.Meta
{
  <doc> Constants for various flags we are using.  </doc>
  const META_STATE_POINTER = 1;
}

<doc> Pass this to the {bottom_type}.  </doc>
OutputStream (s)
  writeCExtensionList OutputStream s
	      forUnit mu.Unit unit
{
  [bottom_type writeCExtensionList s forUnit unit mark ++search_mark];
}

<doc> Pass to the {bottom_type}.  </doc>
MutableArray (collect)
  collectLoadMethods MutableArray collect
	     forUnit mu.Unit unit
{
  [bottom_type collectLoadMethods collect forUnit unit mark ++search_mark];
}

end;

implementation instance
Meta
{
  <doc> Various flags.  </doc>
  public int meta_flags;
}

/******************** mu methods ********************/

<doc> Set the {META_STATE_POINTER} in {meta_flags} if the type of {v}
    is {pointer}.  </doc>
void
  variable MetaVariable v
	in Extension e
{
  if ([v type] == basic_types_from_kind[BT_POINTER])
    meta_flags |= META_STATE_POINTER;
  [super variable v in e];
}

/******************** tesla methods ********************/

void
  allInheritFeatures
{
  [self allInheritFeatures ++search_mark];
}

void
  allInheritFeatures int k
pre
  mark <= k
post
  mark == k
{
  if (mark == k)
    return;
  mark = k;

  /* Get relevant flags from state superclasses.  */
  boolean meta_state_pointer;
  int i, n;
  if (state_supers != nil)
    for ({i = 0; n = [state_supers length];}; i < n; i++)
      {
	Meta m = state_supers[i];
	Meta n = [m posedSelf];
	if (n == self)
	  n = m;
	int mf = [n meta_flags];

	if (mf & META_STATE_POINTER != 0)
	  {
	    meta_state_pointer = TRUE;
	    // break;
	  }
      }
  if (meta_state_pointer)
    meta_flags |= META_STATE_POINTER;

  Meta m = [self posedSelf];
  if (m != self)
    [m allInheritFeatures mark];

  if (behaviour_subs != nil)
    for ({i = 0; n = [behaviour_subs length];}; i < n; i++)
      [behaviour_subs[i] allInheritFeatures mark];
  if (state_subs != nil)
    for ({i = 0; n = [state_subs length];}; i < n; i++)
      [state_subs[i] allInheritFeatures mark];
}

<doc> Tell {all_extensions} that {allUnitLoaded}.  </doc>
void
  allUnitsLoaded
{
  int i, n = [all_extensions length];
  for (i = 0; i < n; i++)
    [all_extensions[i] allUnitsLoaded];
}

<doc> Similar to {bestFormalType}, but now it is known that the {other} is
    an object.  The default implementation returns {nil}.  </doc>
Type
  bestFormalMeta Meta other
{
  = [self isSubclassOf other] ? other : [other isSubclassOf self] ? self : nil;
}

<doc> Ask the {other} the {bestFormalMeta}.  </doc>
Type
  bestFormalType Type other
{
  = self == other ? self : [other bestFormalMeta self];
}

<doc> Clear any posing administration.  </doc>
void
  clearPosing
post
  !posed && !poser
{
  posed = poser = nil;
}

MutableArray (collect)
  collectLoadMethods MutableArray collect
	     forUnit mu.Unit unit
		mark int k
pre
  mark <= k
post
  mark == k
{
  if (mark == k)
    return;
  mark = k;

  int i, n;
  if (behaviour_supers != nil)
    for ({i = 0; n = [behaviour_supers length];}; i < n; i++)
      [behaviour_supers[i] collectLoadMethods collect forUnit unit mark k];

  if (state_supers != nil)
    for ({i = 0; n = [state_supers length];}; i < n; i++)
      [state_supers[i] collectLoadMethods collect forUnit unit mark k];

  for ({i = 0; n = [all_extensions length];}; i < n; i++)
    [all_extensions[i] collectLoadMethods collect forUnit unit];
}

<doc> Append a {TD_REFERENCE}.  </doc>
boolean
  collectTypesDescription MutableByteString description
{
  [description add TD_REFERENCE];
}

<doc> Return 0.  </doc>
protected int (result)
  distanceFromMeta Meta actual
{
  = 0;
}

<doc> Return the direct superclass that starts the only path from the
    receiving meta to the {superclass}.  Return {nil} upon ambiguity
    or nonsense.  </doc>
Meta (direct)
  directSuperFor Meta superclass
{
  boolean b;
  (b, direct) = [self directSuperFor superclass mark ++search_mark];
  = b ? nil : direct;
}

// Should have the postcondition at every traversal method.
// Tue Jul 14 11:48:46 1998, tiggr@gerbil.org
<doc> Hard worker for {directSuperFor}.  The postcondition ensures
    that only one traversal of the metas is done at a time.  </doc>
protected (boolean, Meta) (multiple, direct)
  directSuperFor Meta superclass
	    mark int k
pre
  self != superclass
post
  mark == k
{
  if (mark == k)
    return;
  mark = k;

  Meta found, m;
  boolean b;
  int i, n;

  if (state_supers != nil)
    for ({i = 0; n = [state_supers length];}; i < n; i++)
      {
	Meta s = state_supers[i];

	if (s == superclass)
	  return (FALSE, superclass);

	(b, m) = [s directSuperFor superclass mark k];
	if (b || m != nil)
	  if (!found)
	    found = s;
	  else
	    return (TRUE, nil);
      }

  if (behaviour_supers != nil)
    for ({i = 0; n = [behaviour_supers length];}; i < n; i++)
      {
	Meta s = behaviour_supers[i];

	if (s == superclass)
	  return (FALSE, superclass);

	(b, m) = [s directSuperFor superclass mark k];
	if (b || m != nil)
	  if (!found)
	    found = s;
	  else
	    return (TRUE, nil);
      }

  if (!found && poser != nil)
    {
      if (poser == superclass)
	direct = superclass;
      else
	return [poser directSuperFor superclass mark k];
    }
  else
    direct = found;
}

<doc> Return the length of the path from the {formal} to the receiving
    meta.  </doc>
int
  distanceToFormal Type formal
{
  if (formal == dynamic_type)
    return 0;

  // If this is changed to return the `actual' distance, then every meta
  // should have equal distance to Top and Any.  Otherwise (keeping
  // current semantics), this can be made much faster.
  // Sun May 17 23:56:57 1998, tiggr@gerbil.org
  // = [self distanceToFormal formal mark ++search_mark] >= 0 ? 0 : -1;
  = [formal distanceFromMeta self];
}

<doc> Hard worker for {distanceToFormal}.  </doc>
protected int (result)
  distanceToFormal Type formal
	      mark int k
pre
  k >= mark
{
  if (self == formal)
    return 0;

  if (mark == k)
    return -1;
  mark = k;

  int i, n, d, dist = -2;
  if (state_supers != nil)
    for ({i = 0; n = [state_supers length];}; i < n; i++)
      {
	d = [state_supers[i] distanceToFormal formal mark mark];
	if (d >= 0 && (dist < 0 || d < dist))
	  dist = d;
      }

  if (behaviour_supers != nil)
    for ({i = 0; n = [behaviour_supers length];}; i < n; i++)
      {
	d = [behaviour_supers[i] distanceToFormal formal mark mark];
	if (d >= 0 && (dist < 0 || d < dist))
	  dist = d;
      }

  = dist + 1;
}

<doc> Like {isProperSubclassOf}, but not transitively.  </doc>
boolean
  isDirectSubclassOf Meta meta
{
  = ((state_supers != nil
      && [state_supers indexOfIdentical meta] >= 0)
     || (behaviour_supers != nil
	 && [behaviour_supers indexOfIdentical meta] >= 0));
}

<doc> Like {isSubclassOf}, but do not consider {self} of subclass of
    {self}.  </doc>
boolean
  isProperSubclassOf Meta meta
{
  = self == meta ? FALSE : [self isSubclassOf meta];
}

<doc> Return {TRUE} if {self == meta}.  Otherwise, let the superclasses
    try this and return {TRUE} as soon as one of them does.  </doc>
boolean
  isSubclassOf Meta meta
{
  = [self isSubclassOf meta mark ++search_mark];
}

<doc> Hard worker for {isSubclassOf}.  </doc>
boolean
  isSubclassOf Meta meta
	  mark int k
pre
  k >= mark
{
  if (self == meta)
    return TRUE;
  if (mark == k)
    return FALSE;
  mark = k;

  int i, n;
  if (state_supers != nil)
    for ({i = 0; n = [state_supers length];}; i < n; i++)
      if ([state_supers[i] isSubclassOf meta mark mark])
	return TRUE;
  if (behaviour_supers != nil)
    for ({i = 0; n = [behaviour_supers length];}; i < n; i++)
      if ([behaviour_supers[i] isSubclassOf meta mark mark])
	return TRUE;
}

<doc> Return whether the {meta} is a subclass of {self}.  </doc>
boolean
  isSuperTypeOf Meta meta
{
  = [meta isSubclassOf self];
}

<doc> Return {TRUE} iff the receiving type can be passed as an actual to
    the {formal} type, i.e., if it matches covariantly.  </doc>
boolean
  matchesFormalType id formal
{
  = self == formal || [formal isSuperTypeOf self];
}

<doc> Return a {c} or an {i}.  </doc>
deferred byte
  metaCharacter;

<doc> Return an {m} or a {c}.  </doc>
deferred byte
  metaMetaCharacter;

<doc> Collect methods that fit the {name_parts}, with the given {actuals},
    into {collect} and return that.  If {collect} is {nil}, return a newly
    created array.  If {super_only}, do not search the receiving meta;
    only superclasses.  If, in addition, {single_super} is set, search
    only the specified direct super, issueing an error if the
    {single_super} is not a direct super.  </doc>
MutableArray
   methods MutableArray collect
     named Indexed name_parts
      with Indexed actuals
       for Expression sender
  confined (boolean, Meta) (super_only, single_super)
pre
  !super_only -> !single_super
{
  Meta start = !super_only && poser != nil ? [self posedSelf] : self;
  = [start methods collect named name_parts with actuals for sender
	   confined (super_only, single_super) mark ++search_mark];
}

<doc> Be informed by the extension {ext} that its containing unit has been
    loaded.  </doc>
void
  singleUnitLoaded Extension ext
{
}

<doc> Be informed by the main extension of this meta that the containing
    unit has been loaded (during the first pass).  </doc>
void
  singleUnitLoaded boolean second_pass
pre
  !second_pass
{
  /* Make sure we inherit from Top.  */
  if (![self isSubclassOf top_type])
    [self addBehaviourSuper top_type];

  /* Make sure Any inherits from us (possibly done too often, but that
     does not matter).  */
  if (![bottom_type isSubclassOf self])
    [bottom_type addBehaviourSuper self];
}

<doc> Return the reference type description.  </doc>
TypeDescription
  typeDescription
{
  = object_type_description;
}

deferred OutputStream
  writeCasi OutputStream s;

deferred OutputStream
  writeCisa OutputStream s;

OutputStream (s)
  writeCExtensionList OutputStream s
	      forUnit mu.Unit unit
		 mark int k
pre
  mark <= k
post
  mark == k
{
  if (mark == k)
    return;
  mark = k;

  int i, n;
  if (behaviour_supers != nil)
    for ({i = 0; n = [behaviour_supers length];}; i < n; i++)
      [behaviour_supers[i] writeCExtensionList s forUnit unit mark k];

  if (state_supers != nil)
    for ({i = 0; n = [state_supers length];}; i < n; i++)
      [state_supers[i] writeCExtensionList s forUnit unit mark k];

  for ({i = 0; n = [all_extensions length];}; i < n; i++)
    [all_extensions[i] writeCExtensionList s forUnit unit];
}

OutputStream (s)
  writeCInitializer OutputStream s
	     posing Meta posed
{
  [self writeCMetaDefinitionName [s print "{(void *) &"]];
  [posed writeCMetaDefinitionName [s print ", (void *) &"]];
  [s print "},\n"];
}

<doc> Emit something like {_md_c_tom_All}.  </doc>
OutputStream (s)
  writeCMetaDefinitionName OutputStream s
{
  [self writeCMetaSuffix [s print "_md_"]];
}

<doc> Emit something like {_mr_m_tom_All}.  </doc>
OutputStream (s)
  writeCMetaReferenceName OutputStream s
{
  [self writeCMetaSuffix [s print "_mr_"]];
}

<doc> Emit something like {_mr_c_tom_All}.  </doc>
OutputStream (s)
  writeCReferenceName OutputStream s
{
  [self writeCSuffix [s print "_mr_"]];
}

<doc> Emit something like {i_tom_All}.  </doc>
OutputStream (s)
  writeCReferenceTypeName OutputStream s
{
  [self writeCSuffix s];
}

<doc> Emit something like {c_tom_All} (for an instance).  </doc>
OutputStream (s)
  writeCMetaSuffix OutputStream s
{
  [C write [[C write [s print ([self metaMetaCharacter], "_")]
	       CName [[[main file] unit] name]] print ("_")] CName name];
}

<doc> Emit something like {i_tom_All}.  </doc>
OutputStream (s)
  writeCSuffix OutputStream s
{
  [C write [[C write [s print ([self metaCharacter], "_")]
	       CName [[[main file] unit] name]] print ("_")] CName name];
}

OutputStream (s)
  writeCType OutputStream s
{
  [self writeCSuffix s];
}

void
  writeRCFile OutputStream of
	 pass int pass
      collect MutableArray collect
{
  if (pass == 0)
    {
      /* Define the _md_.  */
      [[self writeCMetaDefinitionName [of print "struct trt_class "]]
	print " = {\n"];

      /* struct trt_class *isa; */
      [[self writeCisa of] print ",\n"];

      /* tom_int asi; */
      [[self writeCasi of] print ",\n"];

      [of print "{\n"];
      {
	/* struct trt_method_dispatch_table *mdt; */
	[of print "0,\n"];

	/* struct trt_extension_offset_table *eot; */
	[of print "0,\n"];

	/* unsigned int num_instances; */
	[of print "0,\n"];

	/* struct name name; */
	[of print ("{\"", name, "\", ", [name length], "},\n")];

	/* struct trt_refvar_offset_table *rvo; */
	[of print "0,\n"];

	/* int instance_size; */
	[of print "0,\n"];

	/* int initial_asi; */
	if (meta_flags & META_STATE_POINTER != 0)
	  [of print "TGC_ASI_EXPLICIT_MARK_P,\n"];
	else
	  [of print "0,\n"];

	/* struct trt_class **class_reference; */
	[[self writeCMetaReferenceName [of print "&"]] print ",\n"];

	/* struct trt_meta_extensions *extensions; */
	[of print "0,\n"];

	/* struct trt_metas *subs, *supers; */
	[of print "0, 0,\n"];

	/* struct trt_class *poser; */
	[of print "0,\n"];

        /* tom_object all_extensions; */
        [of print "0,\n"];

	/* tom_object state_extensions; */
	[of print "0,\n"];

	/* struct trt_class *sole_instance; */
	if ([self isInstance])
	  [of print "0,\n"];
	else
	  [[Meta ([self instance]) writeCMetaDefinitionName [of print "&"]]
	    print ",\n"];

	/* int mark; */
	[of print "0,\n"];

	/* int mark2; */
	[of print "0}\n"];
      }
      [of print "};\n"];

      // Save room for class variables.
      // Wed Jul 15 14:18:11 1998, tiggr@gerbil.org
      if ([self isInstance])
	[[self writeCMetaSuffix [of print "static char __cvars_XXX_"]]
	  print "[100] = {0};\n"];

      /* Define the _mr_.  */
      [[self writeCMetaDefinitionName
	     [[self writeCMetaReferenceName [of print "struct trt_class *"]]
	       print " = &"]] print ";\n"];
    }
}

void
   writeRHFile OutputStream of
	  pass int pass
{
  if (pass == 0)
    {
      /* Declare the typedef for a reference to this meta.  */
      [[self writeCReferenceTypeName [of print "typedef void *"]] print ";\n"];

      /* Declare the reference name for a reference to the class object.  */
      [[self writeCMetaReferenceName [of print "extern struct trt_class *"]]
	print ";\n"];
    }
  else if (pass == 1)
    {
      [[self writeCMetaDefinitionName [of print "extern struct trt_class "]]
	print ";\n"];
    }
}

/******************** private methods ********************/

<doc> Hard worker for {methods named for confined}.  </doc>
protected MutableArray
   methods MutableArray collect
     named Indexed name_parts
      with Indexed actuals
       for Expression sender
  confined (boolean, Meta) (super_only, single_super)
      mark int k
pre
  (k >= mark
   && (!super_only -> !single_super))
{
  /* This code is from
     (OTMMeta.m)methodsNamed:sender:super:confined:list:mark:.  */
  if (mark == k)
    return collect;
  mark = k;

  boolean done = super_only -> !single_super;
  int i, n;

  if (!super_only)
    for ({i = 0; n = [all_extensions length];}; i < n; i++)
      collect = [all_extensions[i] methods collect named name_parts
			       with actuals for sender];

  (done, collect) = [self methods (collect, name_parts, actuals) for sender
			  supers (single_super, state_supers) done done];
  (done, collect) = [self methods (collect, name_parts, actuals) for sender
			  supers (single_super, behaviour_supers) done done];

  if (!done)
    [[[ParseTom error sender] print (single_super, " is not a direct super of ",
				     [sender force_type])] nl];

  = collect;
}

<doc> Hard worker for {methods named confined mark}, visiting the
    indicated {supers}.  </doc>
protected (boolean, MutableArray)
  methods (MutableArray, Indexed, Indexed) (collect, name_parts, actuals)
      for Expression sender
   supers (Meta, Indexed) (single_super, supers)
     done boolean done
{
  if (supers != nil)
    {
      int i, n = [supers length];

      for (i = 0; i < n; i++)
	{
	  Meta m = supers[i];
	  if (!single_super || single_super == m)
	    {
	      Meta posed_m = [m posedSelf];

	      // tomCCC Cast needed.
	      // Tue Apr 28 14:36:23 1998, tiggr@gerbil.org
	      collect = [Meta (self == posed_m ? m : posed_m)
			  methods collect named name_parts with actuals
			  for sender confined (FALSE, nil) mark mark];
	      done = YES;
	    }
	}
    }
  = (done, collect);
}

end; 

implementation class
mu.Class extension tesla: tesla

end;

implementation instance
mu.Class extension tesla

<doc> Return a {c}.  </doc>
byte
  metaCharacter
{
  = 'c';
}

<doc> Return an {m}.  </doc>
byte
  metaMetaCharacter
{
  = 'm';
}

OutputStream (s)
  writeCasi OutputStream s
{
  [s print "TGC_ASI_META_CLASS_P"];
}

OutputStream (s)
  writeCisa OutputStream s
{
  [s print "&_md_m_tom_State"];
}

end;

implementation class
mu.Instance extension tesla: tesla

end;

implementation instance
mu.Instance extension tesla

<doc> Let {super} do its job, and the {meta}.  </doc>
void
  allUnitsLoaded
{
  [super allUnitsLoaded];
  [meta allUnitsLoaded];
}

<doc> Pass to {super} and our {meta}.  </doc>
void
  clearPosing
post
  !posed && !poser
{
  [super clearPosing];
  [meta clearPosing];
}

<doc> Return a {i}.  </doc>
byte
  metaCharacter
{
  = 'i';
}

<doc> Return a {c}.  </doc>
byte
  metaMetaCharacter
{
  = 'c';
}

OutputStream (s)
  writeCasi OutputStream s
{
  [s print "TGC_ASI_CLASS_P"];
}

OutputStream (s)
  writeCisa OutputStream s
{
  [meta writeCMetaDefinitionName [s print "&"]];
}

<doc> Let {super} do its job, and the {meta}.  </doc>
OutputStream (s)
  writeCInitializer OutputStream s
	     posing Instance posed
{
  [super writeCInitializer s posing posed];
  [meta writeCInitializer s posing [posed itsClass]];
}

<doc> Let {super} do its job, and the {meta}.  Also, add {self} to
    {collect}.  </doc>
void
  writeRCFile OutputStream of
	 pass int pass
      collect MutableArray collect
{
  if (pass == 0)
      [collect add self];
  [super writeRCFile of pass pass collect collect];
  [meta writeRCFile of pass pass collect collect];
}

<doc> Let {super} do its job, and the {meta}.  </doc>
void
   writeRHFile OutputStream of
	  pass int pass
{
  [super writeRHFile of pass pass];
  [meta writeRHFile of pass pass];
}

end;
