<copyright> TESLA mu.TomFile extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998, 1999 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: TomFile.t,v 1.24 2000/09/18 06:40:35 bruce Exp $</id>
    </copyright>

implementation class
mu.TomFile extension tesla: tesla
{
  <doc> The extension of source files we emit.  If not set, defaults to
      {.c}.  </doc>
  static mutable String c_file_extension;
}

end;

implementation instance
mu.TomFile extension tesla
{
  <doc> The top-level &lt;c&gt;constructs&lt;/c&gt;.  </doc>
  MutableObjectArray foreign_code;

  <doc> The name of the output file.  Only set if the output needs to be
      written.  </doc>
  String output_filename;

  <doc> Iff {TRUE}, this file only has its interface read, not its
      implementation.  </doc>
  public boolean read_interface;
}

void
  addForeignCode ForeignCode c
{
  if (!foreign_code)
    foreign_code = [MutableObjectArray with c];
  else
    [foreign_code add c];
}

void
  loadFile
{
  if (self == builtin_file)
    {
      [self need_semantics];
      return;
    }

  String read_file = filename;
  if (single_unit != nil && unit != single_unit)
    {
      /* We can do with only the interface file.  */
      MutableByteString fn = [MutableByteString new];
      String bn = [File basename filename without-extension: ".t"];
      read_file = [fn print ([unit directory], bn, ".j")];
      read_interface = TRUE;
    }

  LexTom l = catch (self)
    bind ((file-error,
           {
             [[[ParseTom error] print ("Error opening file \"", read_file,
                                       "\": ", [condition message])] nl];
             [self throw nil];
             nil;
           }))
    ({
      [[LexTom alloc] initWithStream [File open read_file input: YES
                                           flags: FILE_NOT_EXIST_RAISE]
                      named read_file];
    });

  if (!l)
    return;

  ParseTom p = [ParseTom parser self using l];
  if (verbose_parse_file)
    [[[Parser report] print ("reading ", read_file)] nl];
  [p parse];
  [l close];
}

void
  makeExtensionPerform tom.Invocation inv
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [inv fireAt elements[i]];
}

void
  makeMethodsPerform tom.Invocation inv
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [elements[i] makeMethodsPerform inv];
}

<doc> Return {TRUE} iff full semantics of this file need to be read.  This
    is only asked once, just before the parser digests its first token.
    </doc>
boolean
  need_semantics
{
  if (!single_unit || unit == single_unit)
    {
      MutableByteString fn = [MutableByteString new];
      String bn = [File basename filename without-extension: ".t"];
      String ext = !c_file_extension ? ".c" : c_file_extension;

      if (!single_unit)
	[fn print ([unit name], "-", bn, ext)];
      else
	[fn print (bn, ext)];

      // If we would cache information about this file (esp. what happens
      // in the method bodies), we could skip reading it, since we could
      // use the cached information for the RC and RH file.
      // Fri Jul 17 22:36:59 1998, tiggr@gerbil.org
      if (flag_force_output
	  || [File file-needs-updating fn with-respect-to filename])
	output_filename = fn;

      = TRUE;
    }
}

<doc> Pass this to every of our {elements}.  </doc>
void
  singleUnitLoaded boolean second_pass
{
  if (verbose_single_file)
    [[[ParseTom report] print (second_pass ? "2 " : "1 ", filename)] nl];

  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [elements[i] singleUnitLoaded second_pass];
}

<doc> Write the output file, if needed.  If not needed, we act as if we
    write the output file, since the act of emitting the output file
    collects a lot of information that is needed for the RC and RH files.
    </doc>
void
  writeCFile
{
  TeslaOutput of;

  if (output_filename != nil)
    of = [[TeslaOutput alloc] init [File open output_filename output: YES
					 flags: (FILE_NOT_EXIST_CREATE
						 | FILE_EXIST_TRUNCATE)]
			      bufferSize 16384];
  else
    {
      // Maybe TeslaOutput should be an extension of all OutputStreams,
      // including Sink...
      // Fri Jul 17 22:57:25 1998, tiggr@gerbil.org
      of = [[TeslaOutput alloc] init [Sink new]];
    }

  if (!flag_readable_c)
    [of print ("# 1 \"", filename, "\"\n")];
  [[[of print "/* Your mother was a gorilla.  */\n"]
     print ("#define TESLA_OUTPUT\n#include <", [unit name], "-r.h>\n")]
    print "extern void *trt_ext_address (tom_object, int);\n"];

  unwind ([[[of print '\n'] flushOutput] close])
    ({
      int i, n;

      if (foreign_code != nil)
	for ({i = 0; n = [foreign_code length];}; i < n; i++)
	  [[foreign_code[i] precompile] writeCFile of];

      n = [elements length];
      for (i = 0; i < n; i++)
	[Extension (elements[i]) writeCFile of];
    });
}

<doc> Pass to the {elements}.  </doc>
void
  writeRCFile OutputStream of
	 pass int pass
      collect MutableArray collect
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [elements[i] writeRCFile of pass pass collect collect];
}

<doc> Pass to the {elements}.  </doc>
void
  writeRHFile OutputStream of
	 pass int pass
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [elements[i] writeRHFile of pass pass];
}

end; 
