<copyright> TESLA mu.Variable extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Variable.t,v 1.27 2001/04/08 22:39:48 bruce Exp $</id>
    </copyright>

implementation class
Variable extension tesla: tesla
{
  <doc> Various flags.  </doc>
  const CONSTANT_DID_USAGE = 1;
  const VARIABLE_IN_BLOCK_CONTEXT = 2;
}

end;

implementation instance
Variable extension tesla
{
  <doc> Various flags.  </doc>
  int var_flags;
}

<doc> Return the closest enclosing {Callable}.  </doc>
deferred Callable
  callable;

<doc> Do nothing, really.  </doc>
void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
}

<doc> Return {TRUE} if the receiving {Variable} is a instance or class
    variable with a reference type.  The default implementation
    returns {FALSE}.  </doc>
boolean
  isRefICVar
{
}

Cref
    precompile Cblock parent
	method Cblock body
{
  = [[[Ctomvar alloc self] init self] reference nil];
}

<doc> Return a reference to the receiving variable.  If the instance {i}
    is not {nil}, it is the context of the reference (which will be an
    instance).  </doc>
deferred Expression
  reference Instance i;

deferred String
  refICVarOwner;

<doc> Return the first element of the receiving tuple.  The default
    implementation returns {self}.  </doc>
id
  tupleFirstElement
{
  = self;
}

<doc> Emit the declaration in C of the receiving variable.  Default
    implementation does nothing.  </doc>
OutputStream (s)
  writeCDeclaration OutputStream s
{
}

deferred TeslaOutput
  writeCFile TeslaOutput s
      prefix String prefix
       infix String infix
     postfix String postfix
   skipFirst boolean skip_first;

deferred TeslaOutput
  writeCName TeslaOutput s;

TeslaOutput (s)
  writeCReference TeslaOutput s
{
  = [self writeCName s];
}

deferred TeslaOutput
  writeCReturnAssignments TeslaOutput s
		skipFirst boolean skip_first;

end;

/******************** VariableSingle ********************/

implementation class
VariableSingle extension tesla: tesla

end;

implementation instance
VariableSingle extension tesla

<doc> Note that this variable is reference through a block boundary, from
    block {b}.  Inform the declaring context.  </doc>
void
  noteBlockReference BlockDefinition b
{
  if (var_flags & VARIABLE_IN_BLOCK_CONTEXT == 0)
    {
      var_flags |= VARIABLE_IN_BLOCK_CONTEXT;
      [[self callable] noteBlockReference b of self];
    }
}

<doc> Do what is needed to complete the initialization of the context of
    the {parent}.  Default does nothing.  </doc>
void
  precompileContextInitialization Cblock parent
			   method Cblock body
{
}

<doc> Return a reference to the receiving variable.  </doc>
Reference
  reference Instance i
{
  = [[Reference alloc] initWithReferenced self fromInstance i];
}

<doc> Replace the receiving variable with an initialized local variable.
    Unless it is {void}.  </doc>
Variable
  initializeNonArgumentReturnValues Compound parent
{
  if (type == void_type)
    return self;

  LocalVariable lv = [parent localVariable (type, name) for self];
  [parent addExpression [[Assignment alloc] initWithLhs [lv reference nil]
					    rhs [type defaultValue nil]]];
  = lv;
}

Ctomvar
  precompileCompoundDeclaration Cblock parent
{
  if (var_flags & VARIABLE_IN_BLOCK_CONTEXT == 0)
    [parent addDecl [[Ctomvar alloc] init self]];
}

<doc> Default writes nothing.  </doc>
OutputStream (s)
  writeCContextReferencePrefix OutputStream s
{
}

OutputStream (s)
  writeCDeclaration OutputStream s
{
  [self writeCName [[type writeCType s] print ' ']];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
      prefix String prefix
       infix String infix
     postfix String postfix
   skipFirst boolean skip_first
{
  if (!skip_first)
    {
      if (prefix)
	[s print prefix];
      [type writeCType s];
      if (infix)
	[s print infix];
      else
	[s print ' '];
      [self writeCName s];
      if (postfix)
	[s print postfix];
    }
}

OutputStream (s)
  writeCName OutputStream s
{
  [C write s CName name];
}

TeslaOutput (s)
  writeCReturnAssignments TeslaOutput s
		skipFirst boolean skip_first
{
  if (!skip_first)
    [[self writeCName [[self writeCName [[s nl nil] print "*_r"]] print " = "]]
      print ';'];
}

end;

/******************** VariableTuple ********************/

implementation class
VariableTuple extension tesla: tesla

end;

implementation instance
VariableTuple extension tesla

<doc> Return either a {Reference} to the single {Variable} in this tuple,
    or an {ExpressionTuple} of them.  </doc>
Expression
  reference Instance ins
{
  if (length == 1)
    return [self[0] reference ins];

  ExpressionTuple et = [ExpressionTuple withCapacity length];
  int i;

  for (i = 0; i < length; i++)
    [et add [self[i] reference ins]];
  = et;
}

<doc> Pass to the first element.  </doc>
Variable
  tupleFirstElement
{
  = [self[0] tupleFirstElement];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
      prefix String prefix
       infix String infix
     postfix String postfix
   skipFirst boolean skip_first
{
  int i;
  for (i = 0; i < length; i++)
    {
      [self[i] writeCFile s prefix prefix infix infix postfix postfix
	   skipFirst skip_first];
      skip_first = NO;
    }
}

TeslaOutput (s)
  writeCReturnAssignments TeslaOutput s
		skipFirst boolean skip_first
{
  int i;
  for (i = 0; i < length; i++)
    {
      [self[i] writeCReturnAssignments s skipFirst skip_first];
      skip_first = NO;
    }
}

end;

/******************** LocalVariable ********************/

implementation class
LocalVariable extension tesla: tesla

end;

implementation instance
LocalVariable extension tesla

<doc> Pass to the {container}.  </doc>
Callable
  callable
{
  = [container callable];
}

OutputStream (s)
  writeCContextDeclaration OutputStream s
{
  = [self writeCDeclaration s];
}

<doc> Get value from the {callable} context if
    {VARIABLE_IN_BLOCK_CONTEXT}.  </doc>
TeslaOutput (s)
  writeCReference TeslaOutput s
{
  if (var_flags & VARIABLE_IN_BLOCK_CONTEXT != 0)
    [[[self callable] writeContextName s] print "->"];
  = [super writeCReference s];
}

end;

/******************** MetaVariable ********************/

implementation class
MetaVariable extension tesla: tesla

end;

implementation instance
MetaVariable extension tesla

<doc> Return {TRUE} if the receiving {MetaVariable} has a reference
    type.  </doc>
boolean
  isRefICVar
{
  = [type isMeta];
}

<doc> Note the use of the state of this extension if non-static.  </doc>
Cref
    precompile Cblock parent
	method Cbody body
{
  if (![self isStatic])
    [body needExtension ext];
  else if ([self isLocal])
    [body needLocals];

  = [super precompile parent method body];
}

<doc> Return {"self"} if the extension of this meta variable is an
    instance, and so is the {current_meta}.  Return {"self->isa"} if
    we are in a class and the current meta is an instance.  </doc>
String
  refICVarOwner
{
  = [current_meta isInstance] && [[ext meta] isClass]
      ? "(((tom_object) self)->isa)" : "self";
}

<doc> Define the public and mutable methods.  </doc>
void
  singleUnitLoaded boolean second_pass
{
  if (second_pass)
    return;

  // The Qualified class should offer methods to do arithmetic for us.
  // Sun Aug 27 12:24:30 2000, tiggr@home.tiggr.net
  int qualifiers = [self isObsolete] ? 1 << OBSOLETE - EXTERN : 0;

  if ([self isPublic])
    {
      CustomMethod m = [[CustomMethod alloc self]
			 init type qualifiers qualifiers extension ext];
      [m addNamePart name];
      // only fill the body if single_unit
      // Mon May 18 21:01:05 1998, tiggr@gerbil.org
      MethodBody b = [m startMethodBody FALSE];
      ReturnAssignment
	ra = [[ReturnAssignment alloc]
	       initWithLhs [m return_value]
	       rhs [[Reference alloc self]
		     initWithReferenced self fromInstance nil]];
      [b addExpression ra];
      [m finishMethodBody];
    }

  if ([self isMutable])
    {
      CustomMethod m = [[CustomMethod alloc self]
			 init void_type qualifiers qualifiers extension ext];
      String mname = [UniqueString with [[MutableByteString new]
					  print ("set_", name)]];
      String aname = [UniqueString with "_new_value"];
      Argument arg = [[ArgumentSingle alloc self]
		       initWithName aname type type];
      [m addNamePart mname argument: arg];
      MethodBody b = [m startMethodBody FALSE];
      Assignment a = [[Assignment alloc]
		       initWithLhs [[Reference alloc self]
				     initWithReferenced self fromInstance nil]
		       rhs [[Reference alloc self]
			     initWithReferenced arg fromInstance nil]];
      [b addExpression a];
      [m finishMethodBody];
    }
}

<doc> Write the declaration of this meta variable.  </doc>
OutputStream (s)
  writeCDeclaration OutputStream s
{
  boolean is_static = [self isStatic];

  if (is_static && [self isLocal])
    [self writeCName [s print "extern int "]];
  else
    {
      if (is_static)
	{
	  [[type writeCType [s print "extern "]] print ' '];
	  [[ext writeCNameSuffix s] print "_"];
	}
      else
	[[type writeCType s] print ' '];

      [super writeCName s];
    }

  [s print ';'];
  if (is_static)
    [s nl];
}

<doc> If this is a static variable, emit its definition (as in
    {c_tom_Runtime_gc_partial_threshold}), or in case of a
    thread-local class variable, its offset (as in
    {_tlo_c_tom_Thread_current}).  </doc>
MutableArray (a)
  writeDefineStaticVariables TeslaOutput s
		     collect MutableArray a
{
  if ([self isStatic])
    {
      if ([self isLocal])
	[self writeCName [s print "int "]];
      else
	{
	  [[type writeCType s] print ' '];
	  [C write [[ext writeCNameSuffix s] print "_"] CName name];
	}
      [[s print ';'] nl];

      if (!a)
	a = [MutableObjectArray with self];
      else
	[a add self];
    }
}

<doc> Precede with the name suffix of the extension {ext} if this
    variable is static.  Precede with {_tlo_} if thread-local.  </doc>
OutputStream (s)
  writeCName OutputStream s
{
  if ([self isStatic])
    {
      if ([self isLocal])
	[s print "_tlo_"];
      [[ext writeCNameSuffix s] print "_"];
    }
  else
    [ext writeCNameStatePrefix s];
  [super writeCName s];
}

OutputStream
  writeCReference OutputStream s
{
  if (![self isLocal])
    [self writeCName s];
  else
    {
      [type writeCType [s print "*("]];
      [s print " *) (_th_locals + "];
      [self writeCName s];
      [s print ")"];
    }
}

TeslaOutput (s)
  writeCNormalDescription TeslaOutput s
{
  [s print ("\n{{{\"", name, "\", ", [name length])];
  TypeDescription td = [type typeDescription];
  [s print ("}, ", [td trtName], "}, ")];
  [ext writeCExtensionStructName [s print "(char *) &((struct "]];
  [[C write [s print " *) 0)->"] CName name] print " - (char *) 0},"];
}

TeslaOutput (s)
  writeCStaticDescription TeslaOutput s
{
  [s print ("\n{{{\"", name, "\", ", [name length])];
  TypeDescription td = [type typeDescription];
  [self writeCName [s print ("}, ", [td trtName], "}, &")]];
  [s print [self isLocal] ? ", 1}," : ", 0},"];
}

end;

/******************** TemporaryVariable ********************/

implementation class
TemporaryVariable: LocalVariable
{
  <doc> Counter for uniquifying the names of temporaries.  </doc>
  static int next_local_id;
}

<doc> Designated allocator.  </doc>
instance (id)
  withType Type t
	in Compound c
pre
  ![t isTuple]
{
  MutableByteString nm = [[MutableByteString new] print ('_', next_local_id++)];
  = [[self alloc] initWithName [UniqueString with nm] type t container c];
}

<doc> Reset the {next_local_id} counter.  </doc>
void
  reset
{
  next_local_id = 0;
}

end;

implementation instance
TemporaryVariable

<doc> Emit an assignment to 0 and return {self}.  </doc>
Variable
  initializeNonArgumentReturnValues Compound parent
{
  [parent addExpression [[Assignment alloc] initWithLhs [self reference nil]
					    rhs [type defaultValue nil]]];
  = self;
}

end;

/******************** BlockVariable ********************/

implementation class
BlockVariable extension tesla

end;

implementation instance
BlockVariable extension tesla

// This method is obviously nicked from MetaVariable, hence its name.
// Tue Oct  5 20:16:01 1999, tiggr@thesis.gerbil.org
<doc> Emit a description of this block variable to the output file {s}.
    </doc>
TeslaOutput (s)
  writeCNormalDescription TeslaOutput s
{
  [s print ("\n{{{\"", name, "\", ", [name length])];
  TypeDescription td = [type typeDescription];
  [s print ("}, ", [td trtName], "}, ")];
  [block writeBlockStructName [s print "(char *) &((struct "]];
  [[C write [s print " *) 0)->"] CName name] print " - (char *) 0},"];
}

OutputStream
  writeCReference OutputStream s
{
  [self writeCName [[block writeBlockName
			   [self writeCContextReferencePrefix s]] print "->"]];
}

void
  coarseUsage int f
       parent Compound c
         body MethodBody m
{
  [super coarseUsage f parent c body m];
  [initial_value coarseUsage f parent c body m];
}

void
  resolveInitialValue Meta context_type
{
  if (initial_value == nil)
    initial_value = [type defaultValue nil];

  initial_value = [initial_value resolve (nil, nil) context context_type];
}

end;

/******************** VariableBlockProxy ********************/

implementation class
VariableBlockProxy: Variable

end;

implementation instance
VariableBlockProxy
{
  <doc> The variable referenced.  </doc>
  VariableSingle variable;

  <doc> The block from which the variable is reference.  </doc>
  BlockDefinition block;
}

id
  initWithVariable LocalVariable v
	 fromBlock BlockDefinition b
{
  (variable, block) = (v, b);
  [variable noteBlockReference b];

  // Duplication of information!
  // Wed Sep  1 22:33:53 1999, tiggr@thesis.gerbil.org
  = [super initWithType [variable type]];
}

// From VariableSingle...
// Wed Sep  1 22:54:02 1999, tiggr@thesis.gerbil.org
<doc> Return a reference to the receiving variable.  </doc>
Reference
  reference Instance i
{
  = [[Reference alloc] initWithReferenced self fromInstance i];
}

<doc> Ignore.  </doc>
void
  noteBlockReference BlockDefinition b
{
}

OutputStream
  writeCReference OutputStream s
{
  [variable writeCContextReferencePrefix s];
  [variable writeCName [[block writeContextName s] print "->"]];
}

end;
