<copyright> TESLA tesla class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: tesla.t,v 1.42 2001/08/26 22:36:17 bruce Exp $</id>
    </copyright>

implementation class
tesla: mu, stdio, tom.Constants, tom.Conditions,
       tesla-config.tom-host, tesla-config.tom-target
{
  <doc> The types discerned in a type description.  </doc>
  const TD_VOID = 'v';
  const TD_BOOLEAN = 'o';
  const TD_BYTE = 'b';
  const TD_CHAR = 'c';
  const TD_INT = 'i';
  const TD_LONG = 'l';
  const TD_FLOAT = 'f';
  const TD_DOUBLE = 'd';
  const TD_POINTER = 'p';
  const TD_SELECTOR = 's';
  const TD_DYNAMIC = 'x';
  const TD_REFERENCE = 'r';

  <doc> String variants of the above.  </doc>
  const TDS_VOID = "v";
  const TDS_INT = "i";
  const TDS_REFERENCE = "r";

  <doc> The maximum sensible alignment (in bits) a type can require.  </doc>
  const MAX_ALIGNMENT = 64;

  <doc> The {State} class.  </doc>
  static Class state_class;

  <doc> The {Block} instance.  </doc>
  static Instance block_instance;

  <doc> The name of the unit being scrutinized.  </doc>
  static mu.Unit top_unit;

  <doc> Iff set, this is the only unit we are looking at.  </doc>
  static mu.Unit single_unit;

  <doc> The unit of the application that will dload the {top_unit}, or
      {nil} if the {top_unit} is not being dloaded.  </doc>
  static mu.Unit app_unit;

  <doc> The selector of the main method.  </doc>
  static Selector main_selector;

  <doc> The selector of the load methods.  </doc>
  static Selector load_selector;

  <doc> The meta that is currently resolving.  Note that during parsing
      (and in the {ParseTom} tesla extension), the {current_meta} of the
      {TomParser} is significant, and this one here is not unused.  </doc>
  static Meta current_meta;

  <doc> The type descriptions for the basic types, keyed on their kind.
      </doc>
  static Array basic_type_descriptions;

  <doc> A single cons cell with the top type.  </doc>
  static Cons top_type_list;

  <doc> The object reference type description.  </doc>
  static TypeDescription object_type_description;

  <doc> Iff {TRUE}, print the top expressions while they are read.  </doc>
  static boolean verbose_parse_pre;

  <doc> Iff {TRUE}, print the methods after they have been resolved.
      </doc>
  static boolean verbose_post_single;

  <doc> Iff {TRUE}, announce the files being are read.  </doc>
  static boolean verbose_parse_file;

  <doc> Iff {TRUE}, announce each file as it is resolved.  </doc>
  static boolean verbose_single_file;

  <doc> We do not have a unit tom if this is true.  </doc>
  static boolean flag_no_tom_unit;

  <doc> Iff {TRUE}, warn about empty compounds.  </doc>
  static boolean flag_warn_empty_compound;

  <doc> Whether or not the output will be readable.  </doc>
  static boolean flag_readable_c;

  <doc> Whether all output files are to be written, irrespective of there
      existence and not being out-of-date.  </doc>
  static boolean flag_force_output;

  <doc> How much do we say.  </doc>
  static int flag_verbosity;
}

<doc> Initialize.  </doc>
void
  initialize
{
  // I though this was no longer necessary...
  // Thu Apr 23 16:47:44 1998, tiggr@gerbil.org
  // This is necessary, because this is what goes on to init
  // tokens_by_number in the parser.
  // Fri Nov 16 19:31:10 2000, bruce@cubik.org
  [ParseUnit set_trace_default [ParseUnit trace_default]];
  [ParseTom set_trace_default [ParseTom trace_default]];

  [Parser addToPath "."];
  [mu init flag_no_tom_unit];
  [tom-host initialize];
  [tom-target initialize];
  [BuiltinMethod initializeOperators];

  TypeDescription td_r = [TypeDescription withName TDS_REFERENCE];
  TypeDescription td_i = [TypeDescription withName TDS_INT];
  TypeDescription td_v = [TypeDescription withName TDS_VOID];

  [[top_type extensionNamed nil]
    addConst ([UniqueString with "nil"],
	      [[ForeignExpression alloc] initWithExpression "0"
					 type bottom_type])];
  top_type_list = [Cons with (top_type, nil)];

  object_type_description = td_r;

  // Should do a i_main_(ip)...
  // Thu Apr 30 01:16:12 1998, tiggr@gerbil.org
  main_selector = [Selector withName "i_main_r" types (td_r, td_i)];
  load_selector = [Selector withName "v_load_r" types (td_r, td_v)];
}

<doc> Get the ball rolling.  </doc>
int (rc)
  main Array arguments
{
  boolean conditioning;

  bind ((condition,
	 ({
           if ([condition condition_class] == runtime-fatal)
             [Runtime exit 1];
	   if (conditioning)
	     [[err print ("recursive condition: ", condition)] nl];
	   else
	     {
	       conditioning = TRUE;
	       [[err print ("condition ",
			    [condition raised] ? "raised" : "signaled",
			    ": ", condition)] nl];
	       Common e = [condition object];
	       if (!!e)
		 [[err print ("FILE=", [e containing_file],
			      " LINE=", [e line])] nl];
	       conditioning = FALSE;
	     }
	   nil;
	 })))
    ({
      rc = [self real_main arguments];
      [self report [ParseUnit self] errors "unit"];
      [self report [ParseTom self] errors "tom"];
    });
}

<doc> The real thing.  </doc>
int
  real_main Array arguments
{
  String top_unit_name, top_file_name, single_file;
  MutableArray class_names, ext_names;
  int i, n = [arguments length];
  boolean flag_1;

  err = [BufferedStream with err];

  if ([ParseTom trace_default])
    [[[ParseTom report] print "apparently tracing TOM parser"] nl];

  for (i = 0; i < n; i++)
    {
      String s = arguments[i];
      if (s[0] == '-')
	if ([s equal "-1"])
	  flag_1 = TRUE;
	else if ([s equal "-a"])
	  flag_no_tom_unit = TRUE;
	else if ([s equal "-v"])
	  flag_verbosity++;
	else if ([s equal "-ve"])
	  verbose_parse_pre = TRUE;
	else if ([s equal "-no-ve"])
	  verbose_parse_pre = FALSE;
	else if ([s equal "-vf"])
	  verbose_parse_file = TRUE;
	else if ([s equal "-no-vf"])
	  verbose_parse_file = FALSE;
	else if ([s equal "-vF"])
	  verbose_single_file = TRUE;
	else if ([s equal "-no-vF"])
	  verbose_single_file = FALSE;
	else if ([s equal "-vm"])
	  verbose_post_single = TRUE;
	else if ([s equal "-no-vm"])
	  verbose_post_single = FALSE;
	else if ([s equal "--version"])
	  {
	    if (flag_verbosity > 0)
	      [[out print ("tesla ", TESLA_VERSION,
			   ", host=", [tom-host tom-config-name],
			   ", target=", [tom-target tom-config-name])] nl];
	    else
	      [[out print TESLA_VERSION] nl];
	    return 0;
	  }
	else if ([s equal "-F"])
	  flag_force_output = TRUE;
	else if (s[1] == 'W')
	  if ([s equal "-Wempty-compound"])
	    flag_warn_empty_compound = TRUE;
	  else if ([s equal "-Wno-empty-compound"])
	    flag_warn_empty_compound = FALSE;
	  else
	    [self usage s];
	else if ([s equal "-freadable-c"])
	  flag_readable_c = TRUE;
	else if ([s equal "-fno-readable-c"])
	  flag_readable_c = FALSE;
	else if (i < n - 1)
	  if ([s equal "-C"])
	    if (!class_names)
	      class_names
		= [MutableObjectArray with ({String s = arguments[++i];})];
	    else
	      [class_names add ({String s = arguments[++i];})];
	  else if ([s equal "-E"])
	    if (!ext_names)
	      ext_names
		= [MutableObjectArray with ({String s = arguments[++i];})];
	    else
	      [ext_names add ({String s = arguments[++i];})];
	  else if ([s equal "-I"])
	    [Parser addToPath arguments[++i]];
	  else if ([s equal "-f"])
	    top_file_name = arguments[++i];
	  else if ([s equal "-o"])
	    single_file = arguments[++i];
	  else if ([s equal "-u"])
	    top_unit_name = arguments[++i];
	  else if ([s equal "--c-extension"])
	    [TomFile set_c_file_extension arguments[++i]];
	  else
	    [self usage s];
	else
	  [self usage s];
      else
	[self usage s];
    }

  if (!(!top_unit_name ^ !top_file_name))
    [self usage "missing or conflicting -u/-f"];

  [self initialize];

  if (top_file_name != nil)
    {
      /* Construct a unit for the single-file.  */
      if (!class_names)
	class_names = [MutableObjectArray with [File basename top_file_name]];
      top_unit = [mu.Unit withSingleFile top_file_name
		    classes class_names extensions ext_names
		    dependingOn (flag_no_tom_unit ? nil
				 : [ObjectArray with UNIT_NAME_TOM])];
      if (!top_unit)
	return 2;
    }
  else if (class_names != nil || ext_names != nil)
    [self usage "-C/-E conflict with -u"];
  else
    {
      /* Load the .u files.  */
      top_unit = [mu.Unit named top_unit_name load: YES];
      if (!top_unit || [ParseUnit num_errors] > 0)
	return 1;
    }

  if (!flag_no_tom_unit)
    {
      unit_tom = [mu.Unit named UNIT_NAME_TOM];
      if (!unit_tom)
	{
	  [[[ParseUnit error]
	     print ("unit ", UNIT_NAME_TOM, " not loaded")] nl];
	  return 1;
	}
      else
	{
	  state_class = [[unit_tom metaNamed META_NAME_STATE] itsClass];
	  block_instance = Meta ([unit_tom metaNamed BLOCK_CLASS]);
	}
    }

  if ([ParseUnit num_errors] > 0)
    return 1;

  if (flag_1)
    {
      single_unit = top_unit;
      if (top_unit == [mu.Unit builtin_unit])
	[all_user_units add top_unit];
    }

  [[err print ("Number of units: ", [all_units length])] nl];

  /* Load the .t files.  */
  [all_user_units makeElementsPerform selector (void loadFiles)];
  if ([ParseTom num_errors] > 0)
    return 1;

  if (!app_unit)
    {
      CustomMethod main_method = [CustomMethod main];
      if (!main_method && !single_unit)
	{
	  [[[ParseTom error] print "definition of main method not found"] nl];
	  return 1;
	}
    }

  [top_unit setPosers];

  /* Do everything necessary as the first step after all units have been
     loaded.  */
  [all_units makeElementsPerform selector (void allUnitsLoaded)];
  if ([ParseTom num_errors] > 0)
    return 1;

  [top_type allInheritFeatures];

  [Proxy checkProxies];
  if ([ParseTom num_errors] > 0)
    return 1;

  /* We're set.  */
  [top_unit writeCFiles single_file];

  if ([ParseTom num_errors] > 0)
    return 1;
}

<doc> Report the number of errors and warnings for the parser class {pc}.
    </doc>
void
  report class (Parser) pc
  errors String d
{
  int nw = [pc num_warnings];
  int ne = [pc num_errors];

  if (nw != 0 || ne != 0)
    [[[pc report] print (d, ": ", ne, " error", ne == 1 ? "" : "s",
			 ", ", nw, " warning", nw == 1 ? "" : "s")] nl];
}

<doc> Output usage information and exit.  </doc>
void
  usage String opt_arg
{
  if (opt_arg != nil)
    [[err print ("bad option or argument: ", opt_arg)] nl];
  [[err print ("usage: ", [Runtime program_name],
	       " [-u unit | -f file ] [options]
regular options:
 -I dir   add DIR to the include path
 -1       classic (fully dynamic) behaviour
 -F       write all files that can be written, acting as if they are absent
 -o file  only output the FILE (and the header file)
advanced options:
 -C name  add class NAME to the classes in the -f file
 -C a:b   similar to `-C a' but in addition make A pose as B
 -E x.y   add the extension Y of the class X to the -f file
 --c-extension .foo    use `.foo' as the extension of generated source files 
 -a                    don't depend on the presence of unit tom
 -Wx -Wno-x            warn or do not warn about `x'.  Possible warnings are:
                       empty-compound
 -fx -fno-x            include or exclude the `x' feature.  Known features are:
                       readable-c
default options:
 --version     report the version of tesla and exit
 -v            be verbose
debug options:
 -ve      print top expressions while they are read
 -vm      print methods after having been resolved
 -vf      emit names of files that are read
 -vF      emit names of files that are being resolved")] nl];
  [Runtime exit 2];
}

end;

implementation instance tesla end; 
