<copyright> TUG tug class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: tug.t,v 1.5 2000/11/18 02:40:59 bruce Exp $</id>
    </copyright>

implementation class
tug: mu, stdio, tom.Constants, tom.Conditions
{
  <doc> The output file.  </doc>
  static File unit_file;

  <doc> How much do we say.  </doc>
  static int flag_verbosity;
}

<doc> Initialize.  </doc>
void
  initialize
{
  [ClassScan set_trace_default [ClassScan trace_default]];
  [mu init TRUE];
}

<doc> Get the ball rolling, and afterwards report the errors.  </doc>
int (rc)
  main Array arguments
{
  = [self real_main arguments];
  [self report [ParseUnit self] errors "unit"];
  [self report [ParseTom self] errors "tom"];
}

<doc> The real thing.  </doc>
int
  real_main Array arguments
{
  int i, n, arg, argc = [arguments length];
  MutableObjectArray uses_units;
  String unit_name;

  err = [BufferedStream with err];

  for (arg = 0; arg < argc; arg++)
    {
      String s = arguments[arg];
      if (s[0] != '-')
	break;
      if ([s equal "--version"])
	{
	  if (flag_verbosity > 0)
	    [[out print ("tug ", TUG_VERSION)] nl];
	  else
	    [[out print TUG_VERSION] nl];
	  return 0;
	}
      else if ([s equal "-v"])
	flag_verbosity++;
      else if (arg < argc - 1)
	{
	  String next_arg = arguments[++arg];

	  if ([s equal "-u"])
	    unit_name = next_arg;
	  else if ([s equal "-U"])
	    if (!uses_units)
	      uses_units = [MutableObjectArray with next_arg];
	    else
	      [uses_units add next_arg];
	  else
	    [self usage s];
	}
      else
	[self usage s];
    }

  if (!unit_name)
    [self usage "missing -u"];

  [self initialize];

  String filename = [[MutableByteString new] print (unit_name, ".u")];
  unit_file = catch (self)
    bind ((file-error,
           {
             [[[ParseUnit error] print ([[condition object] name],
                                        ": ", [condition message])] nl];
             [self throw nil];
             nil;
           }))
      ({
        [File open filename output: YES
              flags: FILE_NOT_EXIST_CREATE | FILE_EXIST_TRUNCATE];
      });

  if (!unit_file)
   return [ClassScan num_errors] + [ParseUnit num_errors]
          + [ParseTom num_errors];

  [unit_file print ("/* Your mother was a gorilla.  */\n\nunit ",
		    unit_name, '\n')];

  if (uses_units != nil)
    {
      for ({i = 0; n = [uses_units length];}; i < n; i++)
	[unit_file print (!i ? "\nuses " : ", ", String (uses_units[i]))];
      [unit_file print ";\n"];
    }

  for (void; arg < argc; arg++)
    {
      catch (self)
        bind ((file-error,
               {
                 [[[ParseUnit error] print ([[condition object] name],
                                      ": ", [condition message])] nl];
                 [self throw void];
                 nil;
               }))
          ({
            String filename = arguments[arg];
            File f = [File open filename input: YES];
            LexScan l = [[LexScan alloc] initWithStream f named filename];
            ClassScan sc = [[ClassScan alloc] init l];
            [sc scan_file];
            [f close];
          });
    }

  [[unit_file print "\nend;"] nl];
  [unit_file close];

  = [ClassScan num_errors] + [ParseUnit num_errors] + [ParseTom num_errors];
}

<doc> Report the number of errors and warnings for the parser class {pc}.
    </doc>
void
  report class (Parser) pc
  errors String d
{
  int nw = [pc num_warnings];
  int ne = [pc num_errors];

  if (nw != 0 || ne != 0)
    [[[pc report] print (d, ": ", ne, " error", ne == 1 ? "" : "s",
			 ", ", nw, " warning", nw == 1 ? "" : "s")] nl];
}

<doc> Output usage information and exit.  </doc>
void
  usage String opt_arg
{
  if (opt_arg != nil)
    [[err print ("bad option or argument: ", opt_arg)] nl];
  [[err print ("usage: ", [Runtime program_name],
	       " -u unit [options]
regular options:
 -U u          make the <unit> depend on the unit <u>
default options:
 --version     report the version of tug and exit
 -v            be verbose")] nl];
  [Runtime exit 2];
}

end;

implementation instance tug end; 
